import numpy as _np
import cupy as _cp
from ..util import fastlen as _fastlen


def corr(input, axes=(-1, -2), norm=False, returngpu=False, **kwargs):
    """
    simple autocorrelation of input along axes (default: last two) using gpu
    axes: axes to correlate along, defaults to last two
    norm: do normalisation along non correlation axes and normalise for pair count
    returngpu: retrun a cupy array
    """

    axes = sorted([input.ndim + a if a < 0 else a for a in axes])
    fftshape = [_fastlen(2 * input.shape[ax]) for ax in axes]
    dinput = _cp.array(input)
    if norm:
        dinput *= 1 / dinput.mean(axis=[i for i in range(input.ndim) if i not in axes] or None)
    ret = _cp.fft.rfftn(dinput, fftshape)
    ret = _cp.abs(ret) ** 2
    ret = _cp.fft.irfftn(ret, axes=axes)
    ret = _cp.fft.fftshift(ret, axes=axes)[
        tuple(
            (
                Ellipsis,
                *(slice(ps // 2 - input.shape[ax], ps // 2 + input.shape[ax]) for ax, ps in zip(axes, fftshape)),
            )
        )
    ]
    if norm:
        n = corr(_cp.ones(tuple(input.shape[ax] for ax in axes)), returngpu=True)
        ret /= n
        ret[(...,) + (n < 0.9).nonzero()] = _np.nan
    if not returngpu:
        ret = _cp.asnumpy(ret)
        _cp.get_default_memory_pool().free_all_blocks()
    return ret


_div = _cp.ElementwiseKernel("T x, T count", "T res", "if(count>0.5){res = x/count;} else {res = 0;}", "div")


class correlator2d:
    def __init__(self, mask=None, norm=True) -> None:
        """Simple 2D correlator
        Does the 2D autocorrelation divided by the mask autocorrelation.
        mask: mask to use for the autocorrelation or None for all-ones mask
        norm: normalize input by mean over all pixels.
        """

        self._fftshape = None
        if mask is not None:
            mask = _cp.asarray(mask, dtype=_cp.float32)
            self._maskcorr = self._corr(mask)
            self._mask = mask
        else:
            self._mask = None
            self._maskcorr = None
        self._mean = None
        self._n = 0
        self._nvar = None
        self._norm = norm

    def _corr(self, input: _cp.array) -> _cp.array:
        if self._fftshape is None:
            self._fftshape = [_fastlen(2 * s) for s in input.shape]
        ret = _cp.fft.rfft2(input, s=self._fftshape)
        ret = _cp.abs(ret) ** 2
        ret = _cp.fft.irfft2(ret)
        ret = _cp.fft.fftshift(ret, axes=(-2, -1))[
            ...,
            slice(self._fftshape[-2] // 2 - input.shape[-2], self._fftshape[-2] // 2 + input.shape[-2]),
            slice(self._fftshape[-1] // 2 - input.shape[-1], self._fftshape[-1] // 2 + input.shape[-1]),
        ]
        return ret

    def add(self, input: _np.array, mask=None) -> None:
        if input.ndim != 2:
            raise ValueError("input must be 2D.")
        input = _cp.array(input, dtype=_cp.float32, copy=True)
        if self._norm:
            input /= _cp.mean(input)

        if mask is not None:
            input *= _cp.asarray(mask, dtype=_cp.float32)
            maskcorr = self._corr(mask)
        elif self._mask is not None:
            input *= self._mask
            maskcorr = self._maskcorr
        elif self._maskcorr is not None:
            maskcorr = self._maskcorr
        else:
            mask = _cp.ones_like(input)
            maskcorr = self._corr(mask)
            self._maskcorr = maskcorr

        corr = self._corr(input)
        corr = _div(corr, maskcorr)

        if self._mean is None:
            self._mean = corr
            self._n = 1
            self._nvar = _cp.zeros_like(corr)
        else:
            self._n += 1
            delta = corr - self._mean
            self._mean += delta / self._n
            delta2 = corr - self._mean
            self._nvar += delta * delta2

    @property
    def mean(self) -> _np.array:
        return self._mean.get()

    @property
    def n(self) -> int:
        return self._n

    @property
    def std(self) -> _np.array:
        return _cp.sqrt(self._nvar / self._n).get()



import numpy as _np
import numba as _numba
from math import erf, sqrt

try:
    import cupy as _cp
except ImportError:
    _cp = None


def randomphotons(probs, Nphotons, rng=None, dtype=float):
    if not _np.isfinite(Nphotons):
        return probs
    if isinstance(probs, _np.ndarray):
        if rng is None:
            rng = _np.random.default_rng(_np.random.randint(2**63))
        return rng.poisson(probs * (Nphotons / probs.sum(axis=(-2, -1), keepdims=True))).astype(dtype)
    elif _cp is not None and isinstance(probs, _cp.ndarray):
        if rng is None:
            rng = _cp.random.default_rng(_np.random.randint(2**63))
        elif isinstance(rng, _np.random.Generator):
            rng = _cp.random.Generator(rng.integers(high=2**63))
        return rng.poisson(probs * (Nphotons / probs.sum(axis=(-2, -1), keepdims=True))).astype(dtype)


def chargesharing(img, s, ns=0):
    """
    simulates charge sharing between neighboring pixels:
    for each count in the discrete image img, a uniformly random position within the pixel is chosen
    and a Gaussian distribution with sigma s applied to the neighboring pixels.
    charge that leaves the detector is lost.
    """

    if not s > 0:
        if ns == 0:
            return img
        else:
            return img * (1 + ns * _np.random.randn(*img.shape))

    w = int(_np.ceil(2 * s))
    hits = _np.repeat(_np.indices(img.shape).reshape(2, -1), img.astype(int).ravel(), axis=1).T + _np.array((w, w))
    ret = _np.zeros((img.shape[0] + 2 * w, img.shape[1] + 2 * w))
    __putcharges(ret, hits, s, w, ns)
    ret = ret[w:-w, w:-w]
    return ret


@_numba.njit(fastmath=True, cache=True)
def __putcharges(ret, hits, s, w, ns):
    mu = _np.random.rand(len(hits), 2) - 0.5
    k = _np.zeros((2 * w + 1, 2 * w + 1))
    c = 1 / (2 * sqrt(2) * s)
    if ns > 0:
        scale = 1 + _np.random.randn(len(hits)) * ns
    for i in range(len(hits)):
        k[:] = 1 / 4
        mux, muy = mu[i]
        for x in range(-w, w + 1):
            a = (mux - x) / (sqrt(2) * s)
            k[x + w, :] *= erf(c + a) + erf(c - a)
        for y in range(-w, w + 1):
            b = (muy - y) / (sqrt(2) * s)
            k[:, y + w] *= erf(c + b) + erf(c - b)
        hitx, hity = hits[i]
        for x in range(2 * w + 1):
            for y in range(2 * w + 1):
                ret[hitx + y - w, hity + x - w] += scale[i] * k[x, y] if ns > 0 else k[x, y]



import math
import numpy as _np
import numba as _numba
import contextlib


@contextlib.contextmanager
def corrfunction(shape, z, qmax, xcenter=None, ycenter=None):
    """
    CPU based radial Autocorrelation with q correction

    parameters:
    shape (tuple) of inputs in pixels
    z (scalar) distance of detector in pixels
    qmax (scalar): maximum distance
    optional
    xcenter (scalar): position of center in x direction, defaults to shape[0]/2
    ycenter (scalar): position of center in x direction, defaults to shape[1]/2
    returns a function with signature float[:](float[:,:] image) that does the correlation
    """
    xcenter = xcenter or shape[0] / 2.0
    ycenter = ycenter or shape[1] / 2.0
    y, x = _np.meshgrid(_np.arange(shape[1], dtype=_np.float64), _np.arange(shape[0], dtype=_np.float64))
    x -= xcenter
    y -= ycenter
    d = _np.sqrt(x ** 2 + y ** 2 + z ** 2)
    qx, qy, qz = [(k / d * z) for k in (x, y, z)]
    del x, y, d

    def inner(input, qx, qy, qz):
        out = _np.zeros((shape[0] + 10, qmax), dtype=_np.float64)
        for refx in _numba.prange(shape[0]):
            for refy in range(shape[1]):
                qxr = qx[refx, refy]
                qyr = qy[refx, refy]
                qzr = qz[refx, refy]
                refv = input[refx, refy]
                for direction in range(2):
                    dqx = 0
                    x = refx + direction  # dont dx=0 it twice
                    while -qmax <= dqx <= qmax and 0 <= x < input.shape[0]:
                        dqy = 0
                        y = refy
                        while dqy <= qmax and 0 <= y < input.shape[1]:
                            dq = (qx[x, y] - qxr) ** 2 + (qy[x, y] - qyr) ** 2 + (qz[x, y] - qzr) ** 2
                            qsave = int(round(math.sqrt(dq)))
                            if qsave >= qmax:
                                break
                            val = refv * input[x, y]
                            out[refx, qsave] += val
                            y += 1
                        x += -1 + 2 * direction
        return out

    finner = _numba.njit(inner, parallel=True, fastmath=True).compile("float64[:,:](float64[:,:],float64[:,:],float64[:,:],float64[:,:])")

    def corr(input):
        """
        Do the correlation
        """
        if finner is None:
            raise ValueError("already closed, use within with statement")
        input = _np.asarray(input).astype(_np.float64, copy=False)
        if not all(i == s for (i, s) in zip(input.shape, shape)):
            raise ValueError("not the same shape")
        return _np.sum(finner(input, qx, qy, qz), axis=0)

    yield corr
    qx = qy = qz = finner = shape = None
    for x in locals():
        del x



import numpy as _np
import cupy as _cp
from pathlib import Path as _Path
from ..util import rotation as _rotation

code = (_Path(__file__).parent / "sim.cu").read_text()


def _pinned(shape, dtype):
    size = int(_np.prod(shape))
    mem = _cp.cuda.alloc_pinned_memory(size * _np.dtype(dtype).itemsize)
    ret = _np.frombuffer(mem, dtype, size).reshape(shape, order="C")
    return ret


def simulate_gen(
    simobject,
    Ndet,
    pixelsize,
    detz,
    k=None,
    settings="double",
    maximg=_np.inf,
    detangles=(0, 0),
    detoffset=(0, 0),
    returngpu=False,
    *args,
    **kwargs,
):
    """
    returns an array of simulated wavefields
    parameters:
    simobject: a simobject whose get() returns an Nx4 array with atoms in the first and (x,y,z,phase) of each atom in the last dimension
    Ndet: pixels on the detector
    pixelsize: size of one pixel in same unit as simobjects unit (usally um)
    detz: detector distance in same unit as simobjects unit (usally um)
    k: angular wavenumber, if None (default) get from simobject
    settings: string, default: double_ff_nodist can contain
        single: use single precision (else: double precission)
        nf:     use near field
        scale: apply 1/r intensity scaling
        secondorder: use second order in far field approximation (Fresnel)
        nofast: no fast math
        unknown options will be silently ignored.
    maximg: generate StopIteration after maximg images.
    detangles: (theta,phi) angles of detector rotation around origin
    detoffset: (x,y) offset of detector center from origin
    returngpu: return the wavefield on the gpu instead of the cpu

    """
    if k is None:
        k = simobject.k
    if "single" in settings:
        intype, outtype, kernelname = _np.float32, _np.complex64, "wfkernelf"
    else:
        intype, outtype, kernelname = _np.float64, _np.complex128, "wfkerneld"
    options = []
    if "scale" not in settings:
        options += ["-Dnodist"]
    if "nf" in settings:
        options += ["-Dusenf"]
    if "secondorder" in settings:
        options += ["-Dsecondorder"]
    if "nofast" not in settings:
        options += ["--use_fast_math"]

    if _np.size(Ndet) == 1:
        Ndet = [Ndet, Ndet]

    if _np.any(detangles):
        M = _cp.array(_rotation(*detangles, 0), _np.float32)
    else:
        M = _cp.eye(3, dtype=_np.float32)
        options += ["-Dnodetrot"]

    threadsperblock = (16, 16, 1)
    blockspergrid_x = int(_np.ceil(Ndet[0] / threadsperblock[0]))
    blockspergrid_y = int(_np.ceil(Ndet[1] / threadsperblock[1]))
    blockspergrid = (blockspergrid_x, blockspergrid_y)

    module = _cp.RawModule(code=code, backend="nvcc", options=tuple(["--std=c++11", "-O3", "--restrict"] + options))
    kernel = module.get_function(kernelname)

    d_wf = _cp.zeros((Ndet[0], Ndet[1]), dtype=outtype, order="C")
    d_atoms = _cp.zeros((simobject.N, 4), intype, order="C")
    h_atoms = _pinned((simobject.N, 4), intype)

    def _gen():
        h_atoms[:, :3], h_atoms[:, 3:] = simobject.get2()
        d_atoms.set(h_atoms)
        kernel(
            blockspergrid,
            threadsperblock,
            (
                d_wf,
                d_atoms,
                M,
                _np.float32(detoffset[0]),
                _np.float32(detoffset[1]),
                _np.float32(detz),
                _np.float32(pixelsize),
                _np.float32(k),
                int(Ndet[0]),
                int(Ndet[1]),
                int(simobject.N),
            ),
        )
        count = 1
        while True:
            if count == maximg:
                yield _cp.copy(d_wf) if returngpu else d_wf.get()
            elif count > maximg:
                return
            else:
                h_atoms[:, :3], h_atoms[:, 3:] = simobject.get2()
                d_atoms.set(h_atoms)
                ret = _cp.copy(d_wf) if returngpu else d_wf.get()
                kernel(
                    blockspergrid,
                    threadsperblock,
                    (
                        d_wf,
                        d_atoms,
                        M,
                        _np.float32(detoffset[0]),
                        _np.float32(detoffset[1]),
                        _np.float32(detz),
                        _np.float32(pixelsize),
                        _np.float32(k),
                        int(Ndet[0]),
                        int(Ndet[1]),
                        int(simobject.N),
                    ),
                )
                yield ret
            count += 1

    return _gen()


def simulate(
    Nimg,
    simobject,
    Ndet,
    pixelsize,
    detz,
    k=None,
    settings="double",
    verbose=True,
    detangles=(0, 0),
    detoffset=(0, 0),
    returngpu=False,
    *args,
    **kwargs,
):
    """
    returns an array of simulated wavefields
    parameters:
    Nimg: number of wavefields to simulate
    simobject: a simobject whose get() returns an Nx4 array with atoms in the first and (x,y,z,phase) of each atom in the last dimension
    Ndet: pixels on the detector
    pixelsize: size of one pixel in same unit as simobjects unit (usally um)
    detz: detector distance in same unit as simobjects unit (usally um)
    k: angular wavenumber
    settings: string, default: double_ff_nodist can contain
        single: use single precision
        nf:     use near field
        scale: apply 1/r intensity scaling
        secondorder: use second order in far field approximation (Fresnel)
        nofast: no fast math
        unknown options will be silently ignored.
    detangles: (theta,phi) angles of detector rotation around origin
    detoffset: (x,y) offset of detector center from origin
    returngpu: return the wavefield on the gpu instead of
    """
    if _np.size(Ndet) == 1:
        Ndet = [Ndet, Ndet]
    gen = simulate_gen(
        simobject,
        Ndet,
        pixelsize,
        detz,
        k,
        settings=settings,
        maximg=Nimg,
        detangles=detangles,
        detoffset=detoffset,
        returngpu=returngpu,
        *args,
        **kwargs,
    )
    result = _np.empty((Nimg, Ndet[0], Ndet[1]), _np.complex128)
    try:
        for i, img in enumerate(gen):
            if verbose:
                print(i, end=". ", flush=True)
            result[i, ...] = img
        return result
    except KeyboardInterrupt:
        if verbose:
            print("Interrupted")
        return result[:i]



import math as _math
import numpy as _np
import numba as _numba
from scipy.spatial import cKDTree as _cKDTree


def poisson_disc_sample(r, d, N=_np.inf, ndim=3, k=10, method="auto", rng=None):
    """
    Random points with minimum distance in between inside n-sphere
    Parameters:
    r: radius n-sphere
    d: minidmum distance
    ndim: dimension of n-sphere
    N: number of samples
    method: string
        'darts': throw darts and check if darts fit
        'bridson' use bridson algorithm
        'bridson_dense': use bridson algorithm with minimum distance between sample canditates. will result in denser samples
        'auto': will use darts for small N and bridson only for big N
    k: fidelity parameter, should be >10 for good results
    """
    if rng is None:
        rng = _np.random.default_rng(seed=_np.random.randint(2**63))

    ndim = int(ndim)
    if ndim < 1:
        raise ValueError("ndim should be integer >=1")
    if r < 0 or d < 0:
        raise ValueError("r and d should be positive")
    if d == 0:
        if "bridson" in method:
            raise ValueError("bridson needs d>0")
        if not _np.isfinite(N):
            raise ValueError(
                "Zero distance would cause infinite number of spheres. Limit N or set distance >0"
            )
        method = "darts"
    if (method == "auto" and (N > 0.75 * (1.33 * r / d) ** ndim or N > 1e6)) or (
        method == "bridson" or method == "bridson_dense"
    ):
        fixd = method == "bridson_dense"
        cellsize = d / _math.sqrt(ndim)
        grid_shape = int(_np.ceil((2 * r + 2 * d) / cellsize))
        grid = _np.zeros(ndim * [grid_shape], dtype=_np.int64)
        points = _poisson_disc_sample_bridson(
            grid, r, d, ndim, k, fixd=fixd, seed=rng.integers(2**63)
        )
        points = points[(_np.einsum("ij,ij->i", points, points)) < r**2]
        if len(points) > N:
            points = points[rng.choice(range(len(points)), int(N)), :]
    elif method == "darts" or method == "auto":
        points = _poisson_disc_sample_darts(r, d, N, ndim, k=k)
    else:
        raise NotImplementedError(f"method {method} unknown")
    return points


@_numba.njit()
def _poisson_disc_sample_bridson(grid, r, d, ndim=3, k=10, fixd=False, seed=0):
    """
    using the nd-bridson algorithm. the grid has to be preallocated
    """
    if seed != 0:
        _np.random.seed(seed)

    cellsize = d / _math.sqrt(ndim)
    grid_shape = len(grid)
    gstride = (_np.array(grid.strides) / grid.itemsize).astype(_np.int64)
    p = _np.zeros(ndim)
    while _np.sum((p - r - d) ** 2) > r**2:
        p = 2 * r * _np.random.rand(ndim) + d - r
    if grid_shape < 2:
        return p.reshape(1, -1)
    points = _np.zeros((1024, ndim))
    points[0, :] = p[:]
    queue = [0]
    coords = (p / cellsize).astype(_np.int64)
    n = 0
    flatcoord = _np.sum(gstride * coords)
    grid.ravel()[flatcoord] = n + 1

    while len(queue):
        q = points[queue.pop(_np.random.randint(0, len(queue))), :]
        rand = _np.random.randn(k, ndim)
        norm = _np.sum(rand**2, axis=-1) ** (1 / 2)
        rs = (
            _np.ones(k) * d
            if fixd
            else d * (1 + (2**ndim - 1) * _np.random.rand(k)) ** (1 / ndim)
        )
        ps = rand * (rs / norm).reshape(-1, 1) + q
        for ip in range(len(ps)):
            p = ps[ip, :]
            coord = (p / cellsize).astype(_np.int64)
            if not _fits_bridson(p, d, points, grid, coord):
                continue
            else:
                n += 1
                grid.ravel()[_np.sum(gstride * coord)] = n + 1
                if n >= len(points):
                    tmp = _np.zeros((2 * len(points), ndim))
                    tmp[: len(points)] = points
                    points = tmp
                points[n, :] = p[:]
                if _np.sum((p - r - d) ** 2) < (r + 2 * d) ** 2:
                    queue.append(n)
    return points[: n + 1, :] - r - d


@_numba.njit()
def _fits_bridson(p, d, points, grid, coords):
    """
    does p fit?
    """
    if coords[0] >= len(grid) or coords[0] < 0:
        return False
    elif grid.ndim == 2:  # fastpath for ndim==2
        if coords[1] >= grid.shape[1] or coords[1] < 0:
            return False
        for i in range(max(coords[0] - 2, 0), min(coords[0] + 3, grid.shape[0])):
            for j in range(max(coords[1] - 2, 0), min(coords[1] + 3, grid.shape[1])):
                if (
                    grid[i, j] != 0
                    and _np.sum((p - points[grid[i, j] - 1]) ** 2) <= d**2
                ):
                    return False
        return True

    elif grid.ndim > 1:  # recurse in the dimensions
        for i in range(max(coords[0] - 2, 0), min(coords[0] + 3, len(grid))):
            if not _fits_bridson(p, d, points, grid[i, ...], coords[1:]):
                return False
    else:  # last dimension
        for i in range(max(coords[0] - 2, 0), min(coords[0] + 3, len(grid))):
            if grid[i] != 0 and _np.sum((p - points[grid[i] - 1]) ** 2) <= d**2:
                return False
    return True


def _poisson_disc_sample_darts(r, mindistance, N, d=3, m=None, k=10, rng=None):
    """
    throwing darts and checking using ndtree
    """
    if rng is None:
        rng = _np.random.default_rng(seed=_np.random.randint(2**63))
    N = _np.clip(N, 1, (1.35 * r / mindistance) ** d) if mindistance > 0 else N
    if m is None:
        m = max(1, N / 4)
    m = int(min(N, m))
    N = int(N)
    found = 0
    points = _np.zeros((0, d))
    maxtries = k * N / m
    for k in range(int(maxtries)):
        rand = rng.standard_normal((m, d + 2))
        newpoints = (
            rand[:, :d]
            * ((r + mindistance) / _np.sqrt(_np.einsum("ij,ij->i", rand, rand)))[
                :, None
            ]
        )
        tree = _cKDTree(newpoints)
        n = tree.query_ball_tree(tree, mindistance)
        nn = [len(el) for el in n]
        mask = _np.ones(m, bool)
        for i in range(len(n)):
            if nn[i] > 1:
                mask[i] = False
                for j in n[i][1:]:
                    nn[j] -= 1
        newpoints = newpoints[mask, :]
        if len(points) > 0:
            newpoints = newpoints[
                _cKDTree(points, balanced_tree=False).query(
                    newpoints, distance_upper_bound=mindistance
                )[0]
                > mindistance
            ]
        found += _np.sum((_np.einsum("ij,ij->i", newpoints, newpoints)) < r**2)
        points = _np.vstack((points, newpoints))
        if found > N:
            break
    points = points[(_np.einsum("ij,ij->i", points, points)) < r**2]
    if len(points) > N:
        points = points[:N]
    return points



__all__ = ["cpu", "simobj", "cuda", "common", "time", "cutime", "simple"]

from . import cpu, simobj, time, simple
from .common import *

try:
    from . import cuda

    auto = cuda
except ImportError as _e:
    if any(x in str(_e.args[0]).lower() for x in ["cuda", "libcu", "cupy"]):
        import warnings as _w

        _w.warn("cuda error. cuda time dependent simulation not imported. is cuda available and paths set?")
        auto = cpu
    else:
        print(_e)
except Exception as _e:
    print("Import exception on cuda:")
    print(_e)
try:
    from . import cutime

    autotime = cutime
except (AttributeError, ImportError) as _e:
    if any(x in str(_e.args[0]).lower() for x in ["cuda", "libcu", "cupy"]):
        import warnings as _w

        _w.warn("cuda error. cuda time dependent simulation not imported. is cuda available and paths set?")
        autotime = time
    else:
        print(_e)
except Exception as _e:
    print("Import exception on cutime:")
    print(_e)



#!/bin/env python
import IPython
from optparse import OptionParser
import idi.simulation as sim
import numpy as np
from numpy import pi

# input parsing
parser = OptionParser(usage="usage: %prog [options] sphere/sc/fcc/hcp/cuso4")
parser.add_option('-d', '--Ndet', dest='Ndet', type='int', default=512, help="Number of pixels on detector (default: 512)")
parser.add_option('-r', dest='r', type='float', default=50, help="radius in nm (default: 50)")
parser.add_option('-a', dest='a', type='float', default=50, help="lattice constant in A (default: 5)")
parser.add_option('-n', '--Natoms', type='float', dest='Natoms', default=None, help="Number of atoms (default: 1e5)")
parser.add_option('-p', '--pixelsize', type='float', dest='pixelsize', default=75, help="Size of pixels on detector in um (default: 75)")
parser.add_option('-z', '--detz', dest='detz', type='float', default=10, help="z distance of detector in cm(default: 10)")
parser.add_option('-i', '--Nimg', dest='Nimg', type='int', default=100, help="Number of images (default: 100)")
parser.add_option('-e', '-E', dest='E', type='float', default=1500, help="Energy in eV (default: 1500)")
parser.add_option('-c', '--coherent', action='store_false', dest='rndphase', default=True, help="no random phases")
parser.add_option('-o', '--output', dest='outfile', type='string', default='out.npz', help="output file (default: out.npz)")


def cb_Nunitcell(option, opt, value, parser):
    setattr(parser.values, option.dest, [int(x) for x in value.split(',')])


parser.add_option(
    '--Nunitcells',
    type='string',
    action='callback',
    dest='Nunitcells',
    default=None,
    callback=cb_Nunitcell,
    help="grid: nx,ny,nz Number of unitcells in x y and z direction, overrides Natoms (default: use Natoms)",
)
parser.add_option('--anglex', type='float', dest='ax', default=0, help="crystal: Rotation angle in degree (default: 0)")
parser.add_option('--angley', type='float', dest='ay', default=0, help="crystal: Rotation angle in degree (default: 0)")
parser.add_option('--anglez', type='float', dest='az', default=0, help="crystal: Rotation angle in degree (default: 0)")
parser.add_option('-f', '--fixedpositions', action='store_false', dest='rndpos', default=True, help="sphere: no randomly changing positions")
parser.add_option('--nocuda', action='store_false', dest='cuda', default=True, help="dont use cuda")
(options, args) = parser.parse_args()
if len(args) != 1:
    parser.error("incorrect number of arguments. Specify either sphere, gridsc, gridfcc, gridhcp or gridcuso4. Use --help for help")
if options.Nunitcells is not None and (options.Natoms is not None or args[0] == 'sphere'):
    parser.error("Nunitcells is only allowed for grid* and if Natoms is not specified")

simtype = args[0]
outfile = options.outfile
Natoms = int(options.Natoms) if options.Natoms is not None else int(1e5)
Ndet = int(options.Ndet)
detz = options.detz * 1e4  # in um
pixelsize = options.pixelsize  # in um
Nimg = int(options.Nimg)
E = options.E  # in ev
rndphase = options.rndphase
rndpos = options.rndpos
rotangles = np.array([options.ax, options.ay, options.az]) / 180 * pi
k = 2 * pi / (1.24 / E)  # in 1/um

if simtype == 'sphere':
    r = options.r * 1e-3  # in um
    simobject = sim.simobj.sphere(E, Natoms, r)
    simobject.rndPhase = rndphase
    simobject.rndPos = rndpos
else:
    a = options.a * 1e-4  # in um
    N = options.Nunitcells if options.Nunitcells is not None else Natoms
    print(N)
    if simtype == 'sc':
        simobject = sim.simobj.sc(E, N, a, rotangles)
    elif simtype == 'fcc':
        simobject = sim.simobj.fcc(E, N, a, rotangles)
    elif simtype == 'cuso4':
        simobject = sim.simobj.cuso4(E, N, rotangles)
    elif simtype == 'hcp':
        simobject = sim.simobj.hcp(E, N, a, rotangles)
    else:
        raise NotImplementedError("unknown object to simulate")
    if rndpos:
        raise NotImplementedError("rndpos for crsyals not implemented")
    simobject.rndPhase = rndphase
if options.cuda:
    result = sim.cuda.simulate(Nimg, simobject, Ndet, pixelsize, detz, k)
else:
    print("using cpu")
    result = sim.cpu.simulate(Nimg, simobject, Ndet, pixelsize, detz, k)
result = np.square(np.abs(result))
print("saving")
np.savez_compressed(outfile, result=result, settings=(vars(options), args))



import cupy as _cp
import numpy as _np
from pathlib import Path as _Path
from ..util import rotation as _rotation

code = (_Path(__file__).parent / "simtime.cu").read_text()


def simulate(
    simobject,
    Ndet,
    pixelsize,
    detz,
    k,
    c,
    tau,
    pulsewidth,
    detangles=(0, 0),
    settings="mixed",
    threads=None,
    *args,
    **kwargs,
):
    """
    Time dependent simulation with decaying amplitudes.
    simobject: simobject to use for simulation (in lengthunit)
    pixelsize: pixelsize (in lengthunit)
    detz: Detector-sample distance
    k: angular wave number (in 1/lengthunit)
    c: speed of light in (lengthunit/timeunit)
    tau: decay time (in timeunit)
    pulsewidth: FWHM of gaussian exciation pulse (in timeunit)
    pulsedirection: direction of the excitation pulse (x,y,z)
    settings: string, can contain
        double,single,mixed - precision
        nf - for nearfield form
        scale - do 1/r intensity scaling
    first call with new settings might recompile and take a few seconds
    detangles: (theta,phi) angles of detector rotation around origin
    """
    from math import cos, sin

    if "double" in settings:
        nametmp, namesim = "tempsized", "simulated"
        ttype, atype, inouttype = _np.float64, _np.complex128, _np.float64
    elif "single" in settings:
        nametmp, namesim = "tempsizef", "simulatef"
        ttype, atype, inouttype = _np.float32, _np.complex64, _np.float32
    else:  # mixed
        nametmp, namesim = "tempsizedf", "simulatedf"
        ttype, atype, inouttype = _np.float32, _np.complex128, _np.float64
    options = []
    if "nf" in settings:
        options += ["-Dusenf"]
    if "scale" not in settings:
        options += ["-Dnodist"]
    options += [
        "-dc",
        "--std=c++11",
        "--expt-relaxed-constexpr",
        "-O3",
        "--use_fast_math",
    ]

    module = _cp.RawModule(code=code, backend="nvcc", options=tuple(options))
    ftempsize = module.get_function(nametmp)
    fsimulate = module.get_function(namesim)
    N = simobject.N
    if threads is None:
        if N > 1e8:
            threads = 1
        elif N > 1e7:
            threads = 2
        elif N > 1e6:
            threads = 4
        else:
            threads = 8

    if _np.size(Ndet) == 1:
        Ndet = [Ndet, Ndet]

    det = _cp.array(
        _np.array(
            _np.meshgrid(
                pixelsize * (_np.arange(Ndet[0]) - (Ndet[0] / 2)),
                pixelsize * (_np.arange(Ndet[1]) - (Ndet[1] / 2)),
                detz,
            )
        ).T.reshape(-1, 3),
        inouttype,
    )
    if _np.any(detangles):
        M = _cp.array(_rotation(*detangles, 0), inouttype)
        det = det @ M.T

    pdet = _cp.array([i.data.ptr for i in det], _np.uint64)

    tmpout = _cp.zeros(2, _np.int64)
    ftempsize(grid=(1,), block=(1,), args=(N, tmpout))
    _cp.cuda.get_current_stream().synchronize()
    tempsize = int(_np.max(tmpout.get()))
    temp = [_cp.zeros(tempsize // 8 + 1, _np.int64) for i in range(threads)]
    ptemp = _cp.array([i.data.ptr for i in temp], _np.uint64)

    t = [_cp.zeros(N, ttype) for i in range(2 * threads)]
    a = [_cp.zeros(N, atype) for i in range(2 * threads)]
    pt = _cp.array([i.data.ptr for i in t], _np.uint64)
    pa = _cp.array([i.data.ptr for i in a], _np.uint64)
    data = _cp.array(simobject.get(), inouttype, order="C")
    times = _cp.array(_np.random.randn(simobject.N) * (pulsewidth / 2.35), inouttype)
    # _cp.random.seed(_np.random.randint(2**64-1,dtype=_np.uint64))
    # times = (_cp.random.randn(N,dtype=_np.float64)*(pulsewidth/2.35)).astype(inouttype) #cave: this uses different seed as numpy
    output = _cp.zeros(len(det), inouttype)
    poutput = _cp.array([i.data.ptr for i in output], _np.uint64)
    _cp.cuda.get_current_stream().synchronize()
    for start in range(0, len(pdet), threads):
        cthreads = int(min(len(pdet) - start, threads))
        end = start + cthreads
        fsimulate(
            grid=(1,),
            block=(cthreads,),
            args=(
                data,
                times,
                pdet[start:end],
                float(tau),
                float(c),
                float(k),
                int(N),
                int(cthreads),
                pt,
                pa,
                tempsize,
                ptemp,
                poutput[start:end],
            ),
        )
    _cp.cuda.get_current_stream().synchronize()
    return output.get().reshape(Ndet)


def simulate_gen(
    simobject,
    Ndet,
    pixelsize,
    detz,
    k,
    c,
    tau,
    pulsewidth,
    detangles=(0, 0),
    settings="mixed",
    threads=None,
    maximg=_np.inf,
    *args,
    **kwargs,
):
    """
    Time dependent simulation with decaying amplitudes.
    simobject: simobject to use for simulation (in lengthunit)
    pixelsize: pixelsize (in lengthunit)
    detz: Detector-sample distance
    k: angular wave number (in 1/lengthunit)
    c: speed of light in (lengthunit/timeunit)
    tau: decay time (in timeunit)
    pulsewidth: FWHM of gaussian exciation pulse (in timeunit)
    pulsedirection: direction of the excitation pulse (x,y,z)
    settings: string, can contain
        double,single,mixed - precision
        nf - for nearfield form
        scale - do 1/r intensity scaling
    detangles: (theta,phi) angles of detector rotation around origin
    """

    # as each simulation is slow compared to setup, we dont currently do any work ahead and just call simulate()
    i = 0
    while i < maximg:
        yield simulate(
            simobject,
            Ndet,
            pixelsize,
            detz,
            k,
            c,
            tau,
            pulsewidth,
            detangles,
            settings,
            threads,
            *args,
            **kwargs,
        )
        i += 1



import numpy as _np
from ..util import fastlen as _fastlen


def corr(input, axes=(-2, -1), norm=False, fftfunctions=(_np.fft.rfftn, _np.fft.irfftn), **kwargs):
    """
    simple autocorrelation of input along axes (default: last two)
    axes: axes to correlate along, defaults to last two
    norm: do normalisation along non correlation axes and normalise for pair count
    """
    fft, ifft = fftfunctions
    fftshape = [_fastlen(2 * input.shape[ax]) for ax in axes]
    fftaxes = None if set(axes) == {-1, -2} and input.ndim == 2 else axes
    if norm:
        input = input * (1 / input.mean(axis=tuple([i for i in range(input.ndim) if i not in axes]) or None))
    ret = fft(input, fftshape, axes=fftaxes)
    ret = ret * ret.conj()
    # _ne.evaluate('(ret*conj(ret))', out=ret, casting='same_kind')
    ret = ifft(ret, axes=fftaxes)
    ret = _np.fft.fftshift(ret, axes=axes)[
        tuple((Ellipsis, *(slice(ps // 2 - input.shape[ax], ps // 2 + input.shape[ax]) for ax, ps in zip(axes, fftshape))))
    ]

    if norm:
        n = corr(_np.ones(tuple(input.shape[ax] for ax in axes)), fftfunctions=fftfunctions)
        ret /= n
        ret[(...,) + (n < 0.9).nonzero()] = _np.nan
    return ret


class correlator2D:
    def __init__(self, mask=None, norm: bool = True) -> None:
        """Simple 2D correlator
        Does the 2D autocorrelation divided by the mask autocorrelation.
        mask: mask to use for the autocorrelation or None for all-ones mask
        norm: normalize input by mean over all pixels.
        """
        self._fftshape = None
        if mask is not None:
            mask = _np.asarray(mask, dtype=_np.float32)
            self._maskcorr = self._corr(mask)
            self._mask = mask
        else:
            self._mask = None
            self._maskcorr = None
        self._mean = None
        self._n = 0
        self._nvar = None
        self._norm = norm

    def _corr(self, input: _np.ndarray) -> _np.ndarray:
        if self._fftshape is None:
            self._fftshape = [_fastlen(2 * s) for s in input.shape]
        ret = _np.fft.rfft2(input, s=self._fftshape)
        ret = _np.abs(ret) ** 2
        ret = _np.fft.irfft2(ret)
        ret = _np.fft.fftshift(ret, axes=(-2, -1))[
            ...,
            slice(self._fftshape[-2] // 2 - input.shape[-2], self._fftshape[-2] // 2 + input.shape[-2]),
            slice(self._fftshape[-1] // 2 - input.shape[-1], self._fftshape[-1] // 2 + input.shape[-1]),
        ]
        return ret

    def add(self, input: _np.ndarray, mask=None) -> None:
        if input.ndim != 2:
            raise ValueError("input must be 2D.")
        input = _np.array(input, dtype=_np.float32, copy=True)
        if self._norm:
            input /= _np.mean(input)
        if mask is not None:
            input *= mask
            maskcorr = self._corr(mask)
        elif self._mask is not None:
            input *= self._mask
            maskcorr = self._maskcorr
        elif self._maskcorr is not None:
            maskcorr = self._maskcorr
        else:
            mask = _np.ones_like(input)
            maskcorr = self._corr(mask)
            self._maskcorr = maskcorr

        corr = _np.divide(self._corr(input), maskcorr, where=maskcorr > 0.5, out=_np.zeros_like(maskcorr))

        if self._mean is None:
            self._mean = corr
            self._n = 1
            self._nvar = _np.zeros_like(corr)
        else:
            self._n += 1
            delta = corr - self._mean
            self._mean += delta / self._n
            delta2 = corr - self._mean
            self._nvar += delta * delta2

    @property
    def mean(self) -> _np.ndarray:
        return self._mean

    @property
    def n(self) -> int:
        return self._n

    @property
    def std(self) -> _np.ndarray:
        return _np.sqrt(self._nvar / self._n)



from .accum import *
from .filters import *
from .funchelper import *
from .poissondisk import *
from .random import *
from . import h5util
import numba as _numba
import numpy as _np
import scipy.ndimage as _snd
import numexpr as _ne
import itertools as _it
from functools import lru_cache


def radial_profile(data, center=None, calcStd=False, os=1):
    """
    calculates a ND radial profile of data around center. will ignore nans
    calStd: calculate standard deviation, return tuple of (profile, std)
    os: oversample by a factor. With default 1 the stepsize will be 1 pixel, with 2 it will be .5 pixels etc.
    """
    if center is None:
        center = _np.array(data.shape) // 2
    if len(center) != data.ndim:
        raise TypeError("center should be of length data.ndim")
    center = _np.array(center)[tuple([slice(len(center))] + data.ndim * [None])]
    ind = _np.indices(data.shape)
    r = (_np.rint(os * _np.sqrt(((ind - center) ** 2).sum(axis=0)))).astype(int)
    databin = _np.bincount(r.ravel(), (_np.nan_to_num(data)).ravel())
    nr = _np.bincount(r.ravel(), ((~_np.isnan(data)).astype(float)).ravel())
    radialprofile = databin / nr
    if not calcStd:
        return radialprofile

    data2bin = _np.bincount(r.ravel(), (_np.nan_to_num(data**2)).ravel())
    radial2profile = data2bin / nr
    std = _np.sqrt(radial2profile - radialprofile**2)
    return radialprofile, std


def cutnan(array):
    """
    remove full-nan rows and columns of 2d array
    """
    ind0 = ~_np.all(_np.isnan(array), axis=0)
    ind1 = ~_np.all(_np.isnan(array), axis=1)
    return array[ind1, :][:, ind0]


# https://stackoverflow.com/a/29042041
def bin(ndarray, new_shape, operation="sum"):
    """
    bin an ndarray
    ndarray: nd-array
    new_shape: shape to bin to. shape of ndarray has to be integer multiple of new_shape along each dimension
    operation: string. sum, mean, max, or min. operation to use
    """
    ops = ["sum", "mean", "max", "min"]
    operation = operation.lower()
    if operation not in ops:
        raise ValueError("Operation not supported.")
    if ndarray.ndim != len(new_shape):
        raise ValueError("Shape mismatch: {} -> {}".format(ndarray.shape, new_shape))
    compression_pairs = [(d, c // d) for d, c in zip(new_shape, ndarray.shape)]
    flattened = [i for p in compression_pairs for i in p]
    ndarray = ndarray.reshape(flattened)
    for i in range(len(new_shape)):
        op = getattr(ndarray, operation)
        ndarray = op(-1 * (i + 1))
    return ndarray


def rebin(ndarray, n, operation="mean"):
    """
    rebin an ndarray
    parameters:
    ndarray: nd-array
    n: scalar or list, factor to bin with. if scalar, same factor along all dimensions
    operation: string. sum, mean, max, or min. operation to use
    if n doesnt divide arr along one dimensions, last elements of n will silently be dropped
    """
    if not (_np.size(n) == 1 or _np.size(n) == _np.ndim(ndarray)):
        raise ValueError("n should be scalar or of same length as ndarray has dimensions")
    newshape = _np.maximum(_np.array(ndarray.shape) // n, 1)
    return bin(
        _np.copy(ndarray[tuple((slice(None, s) for s in n * newshape))]),
        newshape,
        operation,
    )


def centered_part(arr, newshape):
    """
    Return the center newshape portion of the array.
    """
    newshape = _np.asarray(newshape)
    currshape = _np.array(arr.shape)
    newshape = _np.minimum(newshape, currshape)
    startind = (_np.ceil((currshape - newshape) / 2)).astype(int)
    endind = startind + newshape
    myslice = [slice(startind[k], endind[k]) for k in range(len(endind))]
    return arr[tuple(myslice)]


def list2array(li):
    def _len(e):
        return 1 if _np.isscalar(e) else len(e)

    maxlen = _np.max([_len(e) for e in li])
    return _np.array([_np.pad(e, (0, maxlen - _len(e)), "constant") for e in li])


@_numba.jit()
def find_center(img, mask, x0=0, y0=0, maxr=500, d=60):
    id = _find_center_jit(img, mask.astype(bool), int(x0), int(y0), int(maxr), int(d))
    return _np.unravel_index(id, (2 * d + 1, 2 * d + 1)) - _np.array([d, d])


@_numba.njit(parallel=True)
def _find_center_jit(img, msk, x0, y0, maxr, d):
    sx, sy = img.shape
    out = _np.empty((2 * d + 1, 2 * d + 1))
    for xs in _numba.prange(x0 - d, x0 + d + 1, 1):
        for ys in range(y0 - d, y0 + d + 1, 1):
            cx = sx // 2 + xs
            cy = sy // 2 + ys
            rx = min(cx, sx - cx, maxr)
            ry = min(cy, sy - cy, maxr)
            err = 0
            cn = 0
            for x in range(-rx + 1, 0):
                for y in range(-ry + 1, ry):
                    if msk[cx + x, cy + y] == 1 and msk[cx - x, cy - y] == 1:
                        cn += 1
                        err += abs(img[cx + x, cy + y] - img[cx - x, cy - y])
            out[xs + d, ys + d] = err / cn
    return out.argmin()


@_numba.vectorize(
    [_numba.float64(_numba.complex128), _numba.float32(_numba.complex64)],
    target="parallel",
)
def abs2(x):
    return x.real * x.real + x.imag * x.imag


@_numba.vectorize(
    [_numba.complex128(_numba.complex128), _numba.complex64(_numba.complex64)],
    target="parallel",
)
def abs2c(x):
    return x.real * x.real + x.imag * x.imag + 0j


def fill(data, invalid=None):
    """
    fill invalid values by closest valid value. invalid: mask of invalid values, default: np.isnan(data)
    """
    if invalid is None:
        invalid = _np.isnan(data)
    ind = _snd.distance_transform_edt(invalid, return_distances=False, return_indices=True)
    return data[tuple(ind)]


def photons_localmax(img, E, thres=0.0):
    """
    photonize image.
    First count whole photons.
    Second count fractional/split photons at local maxima if sum of neighbouring pixeles (over thres) is over 0.5
    """
    data = (img * (img > 0)) / E
    photons = _np.floor(data)  # whole photons
    remainder = data - photons
    remainder_ismax = _np.logical_and(remainder == _snd.filters.maximum_filter(remainder, 3), remainder > thres)
    el = _snd.morphology.generate_binary_structure(2, 1)
    photons += _np.rint(_snd.filters.convolve(remainder, el) * remainder_ismax)  # sum over neighbours
    return photons


def photons_simple(img, E, ev_per_adu=3.65, bg=0):
    return _np.rint(((_np.squeeze(_np.array(img)) * ev_per_adu) - bg) / E)


def create_mask(img, lowthres=5, highthres=95, sigma=10, hotpixelstd=5):
    """
    create mask by high/low threshold (in percentile) in gaussian (with sigma) blurred image and morphologically cleaning.
    if hotpixelstd is not None, pixels higher then hotpixelstd times the standard deviation over masked mean will be ignored.
    """
    blured = _snd.gaussian_filter(img, sigma)
    mask = _np.logical_or.reduce(
        (
            blured < _np.nanpercentile(blured, lowthres),
            blured > _np.nanpercentile(blured, highthres),
            _np.isnan(img),
        )
    )
    sel21 = _snd.morphology.generate_binary_structure(2, 1)
    mask_cleaned = _snd.morphology.binary_dilation(mask, sel21, 2)
    mask_cleaned = _snd.morphology.binary_closing(mask_cleaned, sel21, 20)
    sel22 = _snd.morphology.generate_binary_structure(2, 2)
    mask_cleaned = _snd.morphology.binary_dilation(mask_cleaned, sel22, 2)
    mask = _np.logical_or(mask, mask_cleaned)

    if hotpixelstd is not None:
        hotpixel = img > (_np.mean(img[~mask]) + hotpixelstd * _np.std(img[~mask]))
        hotpixel = _snd.morphology.binary_dilation(hotpixel, sel22, 2)
        mask[hotpixel] = True
    return mask


def diffdist(*args):
    """
    returns Euclidean norm next neighbour difference of n coordinates: |diffdist(x,y,z)=diff(x),diff(y),diff(z)|
    """
    accum = 0
    for arg in args:
        accum += _np.diff(arg) ** 2
    return _np.sqrt(accum)


def atleastnd(array, n):
    """
    adds dimensions of length 1 in front of array to get to n dimensions
    """
    if not isinstance(array, _np.ndarray):
        array = _np.array(array)
    return array[tuple((n - array.ndim) * [None] + [...])]


@args2tuple
@lru_cache(maxsize=None)
def fastlen(x, factors=(2, 3, 5, 7, 11)):
    """
    return N>=x conisting only of the prime factors given as factors
    """
    # fmt: off
    fastlens = _np.array((
                            1,    2,    3,    4,    5,    6,    7,    8,    9,    10,   11,
                            12,   14,   15,   16,   18,   20,   21,   22,   24,   25,   27,
                            28,   30,   32,   33,   35,   36,   40,   42,   44,   45,   48,
                            49,   50,   54,   55,   56,   60,   63,   64,   66,   70,   72,
                            75,   77,   80,   81,   84,   88,   90,   96,   98,   99,   100,
                            105,  108,  110,  112,  120,  121,  125,  126,  128,  132,  135,
                            140,  144,  147,  150,  154,  160,  162,  165,  168,  175,  176,
                            180,  189,  192,  196,  198,  200,  210,  216,  220,  224,  225,
                            231,  240,  242,  243,  245,  250,  252,  256,  264,  270,  275,
                            280,  288,  294,  297,  300,  308,  315,  320,  324,  330,  336,
                            343,  350,  352,  360,  363,  375,  378,  384,  385,  392,  396,
                            400,  405,  420,  432,  440,  441,  448,  450,  462,  480,  484,
                            486,  490,  495,  500,  504,  512,  525,  528,  539,  540,  550,
                            560,  567,  576,  588,  594,  600,  605,  616,  625,  630,  640,
                            648,  660,  672,  675,  686,  693,  700,  704,  720,  726,  729,
                            735,  750,  756,  768,  770,  784,  792,  800,  810,  825,  840,
                            847,  864,  875,  880,  882,  891,  896,  900,  924,  945,  960,
                            968,  972,  980,  990,  1000, 1008, 1024, 1029, 1050, 1056, 1078,
                            1080, 1089, 1100, 1120, 1125, 1134, 1152, 1155, 1176, 1188, 1200,
                            1210, 1215, 1225, 1232, 1250, 1260, 1280, 1296, 1320, 1323, 1331,
                            1344, 1350, 1372, 1375, 1386, 1400, 1408, 1440, 1452, 1458, 1470,
                            1485, 1500, 1512, 1536, 1540, 1568, 1575, 1584, 1600, 1617, 1620,
                            1650, 1680, 1694, 1701, 1715, 1728, 1750, 1760, 1764, 1782, 1792,
                            1800, 1815, 1848, 1875, 1890, 1920, 1925, 1936, 1944, 1960, 1980,
                            2000, 2016, 2025, 2048, 2058, 2079, 2100, 2112, 2156, 2160, 2178,
                            2187, 2200, 2205, 2240, 2250, 2268, 2304, 2310, 2352, 2376, 2400,
                            2401, 2420, 2430, 2450, 2464, 2475, 2500, 2520, 2541, 2560, 2592,
                            2625, 2640, 2646, 2662, 2673, 2688, 2695, 2700, 2744, 2750, 2772,
                            2800, 2816, 2835, 2880, 2904, 2916, 2940, 2970, 3000, 3024, 3025,
                            3072, 3080, 3087, 3125, 3136, 3150, 3168, 3200, 3234, 3240, 3267,
                            3300, 3360, 3375, 3388, 3402, 3430, 3456, 3465, 3500, 3520, 3528,
                            3564, 3584, 3600, 3630, 3645, 3675, 3696, 3750, 3773, 3780, 3840,
                            3850, 3872, 3888, 3920, 3960, 3969, 3993, 4000, 4032, 4050, 4096,
                            4116, 4125, 4158, 4200, 4224, 4235, 4312, 4320, 4356, 4374, 4375
    )) # noqa
    # fmt: on
    if factors != (2, 3, 5, 7, 11) or _np.any(_np.greater(x, 4375)):
        # slow fallback
        fastlens = _np.unique(
            [
                _np.prod(_np.array(factors) ** _np.array(i).astype(float))
                for i in _it.product(*(range(int(2 + _np.log(_np.max(x)) / _np.log(k))) for k in factors))
            ]
        )
    return fastlens[_np.searchsorted(fastlens, x)].astype(int)


def split(x, dx, v=None):
    """
    splits an array x into parts dx apart
    x: 1d array
    dx: scalar
    v (optional): splits v along first axis as x would be split
    returns a list of parts of the array, with empty arrays for empty intervals
    """
    sid = _np.argsort(x)
    s = x[sid]
    ids = []
    last = 0
    for upper in _np.arange(0, s[-1], dx):
        last += _np.argmax(s[last:] > upper)
        ids.append(last)
    ids.append(len(s))
    if v is None:
        return [s[start:stop] for start, stop in zip(ids[0:-1], ids[1:])]
    else:
        return [v[sid[start:stop], ...] for start, stop in zip(ids[0:-1], ids[1:])]


def splits(N, sections, return_lists=False):
    """
    split N in sections ranges
    """
    each, remainder = divmod(N, sections)
    splits = _np.array(([0] + remainder * [each + 1] + (sections - remainder) * [each])).cumsum()
    ranges = [range(splits[i], splits[i + 1]) for i in range(sections)]
    if return_lists:
        return [list(i) for i in ranges]
    else:
        return ranges


def shortsci(number, decimals=0):
    """
    short scientific representation of number with variable precision and no plus or leading zero (!) in exponent as string
    """
    if number == 0:
        return "0e0"
    elif not _np.isfinite(number):
        return str(number)
    else:
        exponent = _np.floor(_np.log10(abs(number)))
        base = round(number / (10**exponent), decimals)
        if abs(base) >= 10:
            base = round(base / 10, decimals)
            exponent += 1
        return f"{base:.{decimals}f}e{int(exponent)}"


@_numba.njit
def axisrotation(axis, theta):
    """
    axis and angle to rotation matrix
    """
    u = axis / _np.linalg.norm(axis.astype(_np.float64))
    return (
        _np.cos(theta) * _np.identity(3) + _np.sin(theta) * _np.cross(_np.identity(3), u) + (1 - _np.cos(theta)) * _np.outer(u, u)
    )


@_numba.njit
def rotation(alpha, beta, gamma):
    """
    angles to rotation matrix
    """
    cosa, cosb, cosg = _np.cos(_np.array((alpha, beta, gamma)))
    sina, sinb, sing = _np.sin(_np.array((alpha, beta, gamma)))

    # yaw pitch roll
    M = _np.array(
        [
            [
                cosb * cosg,
                sina * sinb * cosg - cosa * sing,
                cosa * sinb * cosg + sina * sing,
            ],
            [
                cosb * sing,
                sina * sinb * sing + cosa * cosg,
                cosa * sinb * sing - sina * cosg,
            ],
            [-sinb, sina * cosb, cosa * cosb],
        ]
    )
    return M


def angles(rotmatrix):
    """
    rotation matrix to angles
    """
    return (
        _np.arctan2(rotmatrix[2, 1], rotmatrix[2, 2]),
        -_np.arcsin(rotmatrix[2, 0]),
        _np.arctan2(rotmatrix[1, 0], rotmatrix[0, 0]),
    )


def gnorm(x, fwhm, rho, axis=None):
    """
    generalised normal distribution
    """
    s = _np.log(2) * (2 / fwhm) ** rho
    if axis is not None:
        return _np.exp(_ne.evaluate(f"sum(-abs(x)**rho*s), axis={axis})"))
    else:
        return _ne.evaluate("exp(-abs(x)**rho*s)")


def fwhm(X, Y):
    """
    finds the fwhm of y(x) without any interpolation
    """
    id1 = X[_np.argmax(Y > 0.5 * Y.max())]
    id2 = X[-_np.argmax(Y[::-1] > 0.5 * Y.max())]
    return id2 - id1


def arrayfromiter(iterable, count=-1):
    """
    converts an iterable of nd arrays to an n+1-d array
    """
    try:
        if count == -1:
            count = len(iterable)
    except TypeError:
        return _np.array(list(iterable))
    ret = None
    for i, el in enumerate(iterable):
        if ret is None:
            ret = _np.zeros((count, *_np.shape(el)))
        ret[i, ...] = _np.asarray(el)
    return ret


def indmax(array: _np.ndarray) -> tuple:
    """
    unravel_index(nanargmax(array))

    :param array: nd array
    :return: index of maximum
    """
    return _np.unravel_index(_np.nanargmax(array), array.shape)


def indmin(array: _np.ndarray) -> tuple:
    """
    unravel_index(nanargmin(array))

    :param array: nd array
    :return: index of maximum
    """
    return _np.unravel_index(_np.nanargmin(array), array.shape)


def intersect2d(a: _np.ndarray, b: _np.ndarray) -> _np.ndarray:
    """
    finds identical rows in array a,b

    :param a array1
    :param b array2
    :return array of intersecting rows
    """
    intersect = _np.intersect1d(a.view([("", a.dtype)] * a.shape[1]), b.view([("", b.dtype)] * b.shape[1]))
    return intersect.view(a.dtype).reshape(-1, a.shape[1])


def isdir(string):
    """
    Is string an existing dir?
    :param string: path
    :return: abspath or raise NotADirectoryError if is not a dir
    """
    import os

    if os.path.isdir(string):
        return os.path.abspath(string)
    else:
        raise NotADirectoryError(string)


def isfile(string):
    """
    Is string an existing file?
    :param string: path
    :return: abspath or raise FileNotFoundError if is not a dir
    """
    import os

    if os.path.isfile(string):
        return os.path.abspath(string)
    else:
        raise FileNotFoundError(string)


def alignedarray(shape, dtype=_np.float64, alignment=64, zero=False):
    """
    Create an aligned array
    :param shape: shape of the array
    :param dtype: dtype of the array
    :param alignment: alignment in bytes
    :param zero: if True, the array will be zero'd
    :returns: aligned array
    """
    dtype = _np.dtype(dtype)
    nbytes = _np.prod(shape) * dtype.itemsize
    buf = _np.empty(nbytes + alignment, dtype=_np.uint8)
    start_index = -buf.ctypes.data % alignment
    array = buf[start_index : start_index + nbytes].view(dtype).reshape(shape)
    if zero:
        array[:] = 0
    return array


def pixel2q(pixel, E, detz_um, pixelsize_um):
    """
    returns q in reciprocal nm
    E_ev: photon Energy in eV
    detz: detector distance in um
    pixelsize: detector pixelsize in um
    """
    qstep_nm = pixelsize_um / detz_um * (2 * _np.pi) / (1240 / E)
    return pixel * qstep_nm


def q2pixel(q, E, detz, pixelsize):
    """
    return pixel on the detector/in the correlation
    q: q in reciprocal nm
    E_ev: photon Energy in eV
    detz: detector distance in um
    pixelsize: detector pixelsize in um
    """
    qstep_nm = pixelsize / detz * (2 * _np.pi) / (1240 / E)
    return q / qstep_nm


def plot_2d(
    image,
    mask_center=True,
    mask_cross=False,
    crop=(32, 32),
    E=None,
    detz=None,
    pixelsize=75,
    ax=None,
    colorbar=True,
    rings=(),
    cmap="magma",
    title="",
    q_ticks=True,
    **kwargs,
):
    """
    plots an image.
    mask_center (default true): mask the center cross
    only if E(ev), detz(um) and pixelsize(um) are all not None, q circles cam be added and q_ticks can be set
    ax: axis to plot to, if None, gca()
    colorbar: add colorbar
    rings: relative radius of the q circles to add
    other kwargs will be passed to matshow, e.g. vmin/vmax/cmap
    cmap: colormap
    title: title of the plot
    q_ticks: if True, q-ticks will be added, else pixel-ticks
    returns the axis
    """
    import matplotlib.pyplot as plt
    from matplotlib.patches import Circle

    if crop is not None:
        image = centered_part(image, crop)

    if mask_cross:
        i = image.copy()
        centered_part(i, (3, 1))[:] = _np.nan
        centered_part(i, (1, 3))[:] = _np.nan
    elif mask_center:
        i = image.copy()
        centered_part(i, (1, 1))[:] = _np.nan
    else:
        i = image
    if ax is None:
        ax = plt.gca()
    plot = ax.matshow(i, cmap=cmap, **kwargs)
    hasq = not any(x is None for x in (E, pixelsize, detz))
    if hasq:
        maxq = pixel2q(max(i.shape) / 2, E, detz, pixelsize)
        if maxq < 1:
            # swith to 1/um
            qscale = 1000
            qunit = r"$\mu m^{-1}$"
        else:
            qscale = 1
            qunit = r"$nm^{-1}$"
        labeldir = 1
        for r in rings:
            q = _np.round(r * maxq, int(_np.ceil(1 - _np.log10(maxq))))
            p = q2pixel(q, E, detz, pixelsize)
            circ = Circle(
                _np.array(i.shape) // 2,
                p,
                fill=False,
                linestyle=":",
                alpha=1,
                color="white",
            )
            ax.add_patch(circ)
            ax.text(
                i.shape[0] / 2 + p / _np.sqrt(2),
                i.shape[0] / 2 + labeldir * p / _np.sqrt(2),
                f"{q * qscale}{qunit}",
                fontsize="x-small",
                va="center",
                color="white",
            )
            labeldir *= -1

        Nx, Ny = i.shape
        ax.set_xticks([*range(0, Nx - 1, Nx // 4), Nx - 1])
        ax.set_yticks([*range(0, Ny - 1, Ny // 4), Ny - 1])
        if q_ticks and not any(x is None for x in (E, pixelsize, detz)):
            ax.xaxis.set_major_formatter(lambda x, pos: f"{qscale * (pixel2q(x - (Nx // 2), E, detz, pixelsize)):.2G}")
            ax.yaxis.set_major_formatter(lambda y, pos: f"{qscale * (pixel2q(y - (Ny // 2), E, detz, pixelsize)):.2G}")
            ax.set_xlabel(f"({qunit})")
            ax.set_ylabel(f"({qunit})")
        else:
            ax.xaxis.set_major_formatter(lambda x, pos: f"{int(x - Nx // 2)}")
            ax.yaxis.set_major_formatter(lambda y, pos: f"{int(y - Ny // 2)}")

    if colorbar:
        plt.colorbar(plot, ax=ax)
    if title:
        ax.set_title(title)
    return ax


def plot_radialprofile(image, mask_center=True, E=None, detz=None, pixelsize=75, ax=None, **kwargs):
    """
    plots the radial profile of image.
    mask_center (default true): mask the center cross
    if E(ev), detz(um) and pixelsize(um) are all not None, a q-axis will be added
    ax: axis to plot to, if None, gca()
    returns the axis
    """
    import matplotlib.pyplot as plt

    if mask_center:
        i = image.copy()
        centered_part(i, (3, 1))[:] = _np.nan
        centered_part(i, (1, 3))[:] = _np.nan
    else:
        i = image
    if ax is None:
        ax = plt.gca()
    r = radial_profile(i)
    pixel_x = _np.arange(0, len(r))
    ax.plot(pixel_x, r, **kwargs)
    if not any(x is None for x in (E, pixelsize, detz)):
        q_x = pixel2q(pixel_x, E, detz, pixelsize)
        ax2 = ax.twiny()
        ax2.plot(q_x, _np.repeat(_np.nan, len(q_x)))
        ax2.set_xlabel("$nm^{-1}$")
    ax.set_xlabel("px")
    return ax



import numpy as _np
import numexpr as _ne
from ..util import fastlen as _fastlen


class correlator:
    def __init__(self, mask, fftfunctions=(_np.fft.rfftn, _np.fft.irfftn)):
        """
        nd-correlations with constant mask.
        will not normalize between different added images, but each single image.
        optional fftfunctions argument is pair of forward and backward fft function to use (default: numpy default)
        based on http://www.dirkpadfield.com/Home/MaskedFFTRegistrationPresentation.pdf but instead of subtracting correlations of mean, divide by it
        masked Autocorrelation of Image =          (maskedImage C maskedImage)
                                         --------------------------------------------
                                                (mask C Image) * (Image C mask)
                                                         -----------
                                                         (mask C mask)
        """
        fft, ifft = fftfunctions
        self._fft = fftfunctions
        self._shape = mask.shape
        self._padshape = tuple(_fastlen(2 * s) for s in mask.shape)
        self._mask = self._pad(mask)
        self._fmask = fft(self._mask)
        self._mCm = ifft(self._fmask * self._fmask.conj())

    def corr(self, image):
        """
        does a new correlation
        """
        for ds, cs in zip(image.shape, self._shape):
            if ds != cs:
                raise ValueError('data has not expected shape')
        fft, ifft = self._fft
        pimage = self._pad(image * self._mask[tuple((slice(0, s) for s in self._shape))])
        fimg = fft(pimage)
        res = ifft(_ne.evaluate('fimg*conj(fimg)'))  # iCi
        _ne.evaluate('fimg*conj(fmask)', local_dict={'fmask': self._fmask, 'fimg': fimg}, out=fimg)
        norm = ifft(fimg)  # iCm
        _ne.evaluate('conj(fimg)', out=fimg)
        # after fftshift it would be mCi=flip iCm in both directions,
        # fftshift effects first row/column differently, so in the unshifted version a multiplication by mCi is:
        # norm *= _np.roll(norm[tuple(norm.ndim * [slice(None, None, -1)])], norm.ndim * [1], range(0, norm.ndim))
        norm *= ifft(fimg)
        _ne.evaluate('where(norm>1e-5,res/norm*mCm,res*mCm)', out=res, local_dict={'res': res, 'norm': norm, 'mCm': self._mCm})
        res = _np.fft.fftshift(res)[tuple((slice(ps // 2 - s + 1, ps // 2 + s) for s, ps in zip(self._shape, self._padshape)))]
        return res

    def _pad(self, data):
        """
        pads data to size, data will be in top left corner of return
        """
        ret = _np.zeros(self._padshape, _np.float64)
        ret[tuple((slice(0, s) for s in self._shape))] = data
        return ret

    @property
    def shape_input(self):
        """
        expected input shape for add function
        """
        return self._shape

    @property
    def shape_result(self):
        """
        shape of the result
        """
        return tuple((2 * s - 1 for s in self._shape))

    @property
    def mask(self):
        """
        used mask
        """
        return self._mask[tuple((slice(0, s) for s in self._shape))]



#!/bin/env python
import numpy as np
from optparse import OptionParser
import numba
from idi.reconstruction.ft import corr
from idi.simulation.common import randomphotons
from idi.util import *

pmax = 8  # todo
numba.config.NUMBA_NUM_THREADS = pmax

parser = OptionParser(usage="usage: %prog [options] inputfile")
parser.add_option('-N', '--Nphotons', dest='Nphotons', type='float', default=1e3,
                  help="Number of Photons (default 1e3)")
parser.add_option('-o', '--output', dest='outfile', type='string', default='',
                  help="output file (default: (inputfile)-(Nphotons)-out.npz)")
parser.add_option('-p', '--pmax', dest='pmax', type='int', default=8,
                  help="max. number of parallel processes (default 8)")
parser.add_option('-b', '--bin', dest='bin', type='int', default=0,
                  help="bin 4^n pixels (default 0)")

(options, args) = parser.parse_args()
if len(args) != 1:
    parser.error("incorrect number of arguments. Specify inputfile")
infile = args[0]
Nphotons = int(options.Nphotons)
pmax = options.pmax
nrebin = options.bin
outfile = (
    '%s-%i-out.npz' % (infile if nrebin == 0 else infile + '-b%i' % nrebin, Nphotons)
    if options.outfile == ''
    else options.outfile
)

incoherent = np.load(infile)['result']
incoherent = np.square(np.abs(incoherent))
simsettings = np.load(infile)['settings']
simtype = simsettings[1]
simsettings = simsettings[0]
pixelsize = simsettings['pixelsize'] * 1e-6  # in m
pixelsize = pixelsize * 2 ** nrebin
dz = simsettings['detz'] / 100  # in m
z = dz / pixelsize

photons = np.asarray(randomphotons(rebin(incoherent, nrebin), Nphotons)).astype(np.uint32)
out = corr(photons, z)
photonsum = np.sum(photons, axis=0)
np.savez_compressed(
    outfile,
    corr=out,
    photonsum=photonsum,
    simsettings=vars(simsettings),
    simtype=simtype,
    reconsettings=(vars(options), args),
)



import numpy as _np
import numexpr as _ne
from ..util import rebin as _rebin


def simulate(simobject, Ndet, pixelsize, detz, k, fftfunction=_np.fft.fft2):
    """
    Simple (fft based) Fraunhofer small angle approximation
    """
    if _np.size(Ndet) == 1:
        Ndet = [Ndet, Ndet]
    Ndet = _np.array(Ndet)
    dq = k * pixelsize / detz
    maxq = Ndet / 2 * dq
    dx = _np.pi / maxq
    obj = simobject.getImg(dx)
    os = _np.ceil(_np.array(obj.shape) / Ndet).astype(int)
    img = _np.fft.fftshift(fftfunction(obj, s=os * Ndet))
    return _rebin(img, os, 'mean')


def simulate3d(simobject, Ndet, pixelsize, detz, k, fftfunction=_np.fft.fft2):
    """
    Simple (fft based, 3d sliced) Fraunhofer small angle approximation
    """
    if _np.size(Ndet) == 1:
        Ndet = [Ndet, Ndet]
    Ndet = _np.array(Ndet)
    dq = k * pixelsize / detz
    maxq = Ndet / 2 * dq
    dx = _np.pi / maxq
    dz = min(dx)
    obj = simobject.getImg(dx.tolist() + [dz], 3)
    os = _np.ceil(_np.array(obj.shape[:2]) / Ndet).astype(int)
    qz = 1 - _np.sqrt(
        1 - sum(q ** 2 for q in _np.meshgrid(*[(pixelsize / (detz * o)) * _np.arange(-n * o / 2, n * o / 2) for n, o in zip(Ndet, os)]))
    )
    f = _np.fft.fftshift(fftfunction(obj, s=os * Ndet, axes=(0, 1)), axes=(0, 1))
    # img = sum(f[..., i] * _np.exp((-1j * dz * k * i) * qz) for i in range(f.shape[-1]))
    i = _np.arange(0, f.shape[-1])
    img = _ne.evaluate('sum(f * exp(-(1j * dz * k * i) * qz), axis=2)')
    return _rebin(img, os, 'mean')



from functools import wraps as _wraps, reduce as _reduce


def aslist(fn=None):
    def aslist_return(fn):
        @_wraps(fn)
        def aslist_helper(*args, **kw):
            return list(fn(*args, **kw))

        return aslist_helper

    if fn is None:
        return aslist_return
    return aslist_return(fn)


def asgen(fn=None):
    def asgen_return(fn):
        @_wraps(fn)
        def asgen_helper(arg, *args, **kw):
            for item in arg:
                yield fn(item, *args, **kw)

        return asgen_helper

    if fn is None:
        return asgen_return
    return asgen_return(fn)


def parallel(fn=None):
    from collections import deque
    import time

    try:
        from pathos.multiprocessing import Pool
    except ImportError:
        import warnings

        warnings.warn("no pathos available, be careful with parallel decorator and pickling errors.")
        from multiprocessing import Pool

    def parallel_return(fn):
        @_wraps(fn)
        def parallel_helper(arg, *args, **kw):
            with Pool(4) as p:
                q = deque()
                for item in arg:
                    if len(q) > 4 and q[0].ready():
                        yield q.popleft().get()
                    while len(q) > 8 and not q[0].ready():
                        time.sleep(0.01)
                    q.append(p.apply_async(fn, (item,) + args, kw))
                for r in q:
                    yield r.get()

        return parallel_helper

    if fn is None:
        return parallel_return
    return parallel_return(fn)


def chain(*fns):
    return _reduce(lambda f, g: lambda x: f(g(x)), fns)


def group(iterable, n):
    return zip(*([iter(iterable)] * n))


def args2tuple(function):
    from numpy import ndarray

    def wrapper(*args, **kwargs):
        args = [tuple(x) if isinstance(x, list) else x for x in args]
        kwargs = {k: tuple(x) if isinstance(x, list) else x for k, x in kwargs.items()}
        args = [tuple(x.ravel()) if isinstance(x, ndarray) else x for x in args]
        kwargs = {k: tuple(x.ravel()) if isinstance(x, ndarray) else x for k, x in kwargs.items()}
        return function(*args, **kwargs)

    return wrapper



from numpy import pi
import numpy as _np
import numexpr as _ne


def rndgennorm(mu, fwhm, rho, N, rng=None):
    """
    samples from the generalised random normal distribution with given shape parameter rho, mean mu and fwhm.
    for rho=2 this is a normal distribution, for higher rho it approaches a uniform distribution between -fwhm/2 and fwhm/2.
    mu, rho, fwhm must be broadcastable to the number of samples N.
    """
    if rng is None:
        rng = _np.random.default_rng()
    if _np.isscalar(N):
        if not _np.isscalar(fwhm):
            N = (N, len(fwhm))
        elif not _np.isscalar(rho):
            N = (N, len(rho))
    rho = _np.asarray(rho)
    # https://sccn.ucsd.edu/wiki/Generalized_Gaussian_Probability_Density_Function
    # https://en.wikipedia.org/wiki/Generalized_normal_distribution
    # ret=mu + fwhm / 2 * (rng.gamma(1 / rho, 1, N) / _np.log(2)) ** (1 / rho) * rng.choice((-1, 1), N)
    c = rng.choice(_np.array([-1, 1], _np.int8), N)  # noqa
    ret = rng.gamma(1 / rho, 1, N)
    _ne.evaluate('mu + fwhm / 2 * (ret / log(2)) ** (1 / rho) * c', out=ret)
    return ret


def rndstr(N):
    """
    random string of N lower case ascii letters and numbers
    """
    import random
    import string

    return ''.join(random.choices(string.ascii_lowercase + string.digits, k=N))


def rndSphere(r, N, rng=None):
    if rng is None:
        rng = _np.random.default_rng()
    ret = _np.zeros((N, 3))
    t = rng.uniform(-1, 1, N)
    p = rng.uniform(0, 2 * pi, N)
    u = rng.uniform(0, 1, N)
    _ne.evaluate('r*u**(1/3)', out=u)
    _ne.evaluate('u * t', out=ret[:, 0])
    _ne.evaluate('sqrt(1 - t**2)', out=t)
    _ne.evaluate('u * t * sin(p)', out=ret[:, 1])
    _ne.evaluate('u * t * cos(p)', out=ret[:, 2])
    return ret


def rndConvexpoly(N, n, x, rotmatrix, r, ratio=2):
    if rotmatrix is not None:
        n = _np.matmul(rotmatrix, n.T, order='F').T
        x = _np.matmul(rotmatrix, x.T, order='F').T
    found = 0
    ret = []
    criterion = _np.einsum('ij,ij->i', n, x)
    while found < N:
        points = rndSphere(1, int(_np.clip(ratio * (N - found), 1024, N)))  # steps smalller 1024 are just wastefull, limit memory usage
        inside = _np.ones(len(points), bool)
        for i in range(len(n)):  # loop instead of full vectorisation to decrease memory allocation
            inside &= (points @ n[i]) > criterion[i]
        # inside=_np.all(_np.squeeze(_np.matmul(points[None,...],n[...,None]))>_np.einsum('ij,ij->i',n,x)[...,None],axis=0)
        points = points[inside] * r
        found += len(points)
        ret.append(points)
    return _np.concatenate(ret)[: int(N)]


def rndIcosahedron(r, N, rotmatrix=None):

    phi = (1 + _np.sqrt(5)) / 2
    # surface normals
    n = _np.array(
        [
            [0, -1, 2 - phi],
            [0, -1, phi - 2],
            [2 - phi, 0, -1],
            [phi - 2, -0, -1],
            [phi - 2, 0, 1],
            [2 - phi, 0, 1],
            [0, 1, phi - 2],
            [0, 1, 2 - phi],
            [1, phi - 2, 0],
            [1, 2 - phi, 0],
            [-1, phi - 2, 0],
            [-1, 2 - phi, 0],
            [1, -1, -1],
            [-1, -1, -1],
            [1, -1, 1],
            [-1, -1, 1],
            [1, 1, 1],
            [-1, 1, 1],
            [1, 1, -1],
            [-1, 1, -1],
        ]
    )
    # point on each surface for r=1
    x = _np.array(
        [
            [0, 1, -phi],
            [0, 1, phi],
            [0, 1, phi],
            [0, 1, phi],
            [0, 1, -phi],
            [0, 1, -phi],
            [0, -1, phi],
            [0, -1, -phi],
            [-1, phi, 0],
            [-1, -phi, 0],
            [1, phi, 0],
            [1, -phi, 0],
            [0, 1, phi],
            [0, 1, phi],
            [0, 1, -phi],
            [0, 1, -phi],
            [0, -1, -phi],
            [0, -1, -phi],
            [0, -1, phi],
            [0, -1, phi],
        ]
    ) * (1 / _np.sqrt(1 + phi ** 2))

    return rndConvexpoly(N, n, x, rotmatrix, r)


def random_rotation(rng=None, amount=1):
    if rng is None:
        rng = _np.random.default_rng()
    # https://doc.lagout.org/Others/Game%20Development/Programming/Graphics%20Gems%203.pdf
    theta, phi, z = rng.uniform(0, (2.0 * pi * amount, 2.0 * pi, amount))
    V = (_np.cos(phi) * _np.sqrt(z), _np.sin(phi) * _np.sqrt(z), _np.sqrt(1.0 - z))
    sint, cost = _np.sin(theta), _np.cos(theta)
    R = _np.array(((cost, sint, 0), (-sint, cost, 0), (0, 0, 1)))
    M = _np.dot(2 * _np.outer(V, V) - _np.eye(3), R)
    return M


def rndtruncexp(lam, b, N, rng=None):
    '''
    Draw from a truncated exponential distribution
    '''

    if rng is None:
        rng = _np.random.default_rng()
    r = rng.random(N)
    _ne.evaluate('-log1p(r*expm1(-b/lam))*lam', out=r)
    return r
